/*!
 * Copyright (C) 2001-2002 by egnite Software GmbH. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgement:
 *
 *    This product includes software developed by egnite Software GmbH
 *    and its contributors.
 *
 * THIS SOFTWARE IS PROVIDED BY EGNITE SOFTWARE GMBH AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL EGNITE
 * SOFTWARE GMBH OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 */

/*!
 * $Log: httpserv.c,v $
 * Revision 1.1  2002/08/09 18:27:18  harald
 * Renamed to avoid conficts with lib name
 *
 * Revision 1.8  2002/06/26 17:29:00  harald
 * First pre-release with 2.4 stack
 *
 * Revision 1.7  2002/06/12 10:55:29  harald
 * *** empty log message ***
 *
 * Revision 1.6  2002/06/04 19:11:06  harald
 * *** empty log message ***
 *
 * Revision 1.5  2002/05/11 20:59:22  harald
 * Using new standard output
 *
 * Revision 1.4  2002/05/08 16:02:29  harald
 * First Imagecraft compilation
 *
 */

/*!
 * \example httpd/httpd.c
 *
 * Simple multithreaded HTTP daemon.
 */

#include <string.h>

#include <dev/nicrtl.h>
#include <dev/uartavr.h>

#include <sys/thread.h>
#include <sys/timer.h>
#include <sys/print.h>
#include <sys/heap.h>

#include <netinet/sostream.h>
#include <arpa/inet.h>

#include <pro/httpd.h>

#ifdef NUTDEBUG
#include <sys/osdebug.h>
#include <net/netdebug.h>
#endif

#include "ide.h"
#include "fat32.h"


static char *states[] = { "TRM", "<FONT COLOR=#CC0000>RUN</FONT>", "<FONT COLOR=#339966>RDY</FONT>", "SLP" };

static BYTE aSectorBuffer[IDE_SECTOR_SIZE];


/*
 * CGI Sample. See httpd.h for REQUEST structure.
 */
static int ShowQuery(NUTDEVICE *sostream, REQUEST *req)
{
    static const char icca[] =  "<HTML><HEAD><TITLE>Show Query</TITLE></HEAD><BODY>";

    NutHttpSendHeaderTop(sostream, req, 200, "Ok");
    NutHttpSendHeaderBot(sostream, "text/html", -1);

    NutPrintString_P(sostream, icca);

    if(req->req_query)
        NutPrintString(sostream, req->req_query);
    else
        NutPrintString(sostream, "None");
    NutPrintString(sostream, "</BODY>"
                             "</HTML>");
    NutPrintFlush(sostream);

    return 0;
}

/*
 * CGI Sample.
 */
static int ShowThreads(NUTDEVICE *sostream, REQUEST *req)
{
    NUTTHREADINFO *tdp = nutThreadList;
    static const char icca[] = "<HTML><HEAD><TITLE>Show Threads</TITLE></HEAD><BODY>\r\n";
    static const char iccb[] = "<TABLE BORDER><TR><TH>Handle</TH><TH>Name</TH><TH>Priority</TH><TH>Status</TH><TH>Event<BR>Queue</TH><TH>Timer</TH><TH>Stack-<BR>pointer</TH><TH>Free<BR>Stack</TH></TR>\r\n";
    static const char iccc[] = "</TABLE></BODY></HTML>";

    NutHttpSendHeaderTop(sostream, req, 200, "Ok");
    NutHttpSendHeaderBot(sostream, "text/html", -1);

    NutPrintString_P(sostream, icca);

    NutPrintString_P(sostream, iccb);
    while(tdp) {
        NutPrintFormat(sostream, "<TR><TD>%04X</TD><TD>%s</TD><TD>%u</TD><TD>%s</TD><TD>%04X</TD><TD>%04X</TD><TD>%04X</TD><TD>%u</TD><TD>%s</TD></TR>\r\n", 
                        tdp, tdp->td_name, tdp->td_priority,
                        states[tdp->td_state], tdp->td_queue, tdp->td_timer,
                        tdp->td_sp, (u_short)tdp->td_sp - (u_short)tdp->td_memory,
                        *((u_long *)tdp->td_memory) != DEADBEEF ? "Corr" : "OK");
        tdp = tdp->td_next;
    }
    NutPrintString_P(sostream, iccc);
    NutPrintFlush(sostream);

    return 0;
}

/*
 * CGI Sample.
 */
static int ShowTimer(NUTDEVICE *sostream, REQUEST *req)
{
    NUTTIMERINFO *tnp;
    u_long ticks_left;
    static const char icca[] = "<HTML><HEAD><TITLE>Show Threads</TITLE></HEAD><BODY>";
    static const char iccb[] = "<TABLE BORDER><TR><TH>Handle</TH><TH>Countdown</TH><TH>Tick Reload</TH><TH>Callback<BR>Address</TH><TH>Callback<BR>Argument</TH></TR>\r\n";
    static const char iccc[] = "</TABLE></BODY></HTML>";

    NutHttpSendHeaderTop(sostream, req, 200, "Ok");
    NutHttpSendHeaderBot(sostream, "text/html", -1);

    NutPrintString_P(sostream, icca);

    if((tnp = nutTimerList) != 0) {
        NutPrintString_P(sostream, iccb);
        ticks_left = 0;
        while(tnp) {
            ticks_left += tnp->tn_ticks_left;
            NutPrintFormat(sostream, "<TR><TD>%04X</TD><TD>%lu</TD><TD>%lu</TD><TD>%04X</TD><TD>%04X</TD></TR>\r\n", 
                            tnp, ticks_left, tnp->tn_ticks, 
                            tnp->tn_callback, tnp->tn_arg);
            tnp = tnp->tn_next;
        }
    }

    NutPrintString_P(sostream, iccc);
    NutPrintFlush(sostream);

    return 0;
}

/*
 * CGI Sample.
 */
static int ShowSockets(NUTDEVICE *sostream, REQUEST *req)
{
    extern TCPSOCKET *tcpSocketList;
    TCPSOCKET *ts;
    static const char icca[] = "<HTML><HEAD><TITLE>Show Threads</TITLE></HEAD><BODY>";
    static const char iccb[] = {"<TABLE BORDER><TR><TH>Handle</TH><TH>Type</TH><TH>Local</TH><TH>Remote</TH><TH>Status</TH></TR>\r\n"};

    NutHttpSendHeaderTop(sostream, req, 200, "Ok");
    NutHttpSendHeaderBot(sostream, "text/html", -1);

    NutPrintString_P(sostream, icca);

    NutPrintString_P(sostream, iccb);
    for(ts = tcpSocketList; ts; ts = ts->so_next) {
        NutPrintFormat(sostream, "<TR><TD>%04X</TD><TD>TCP</TD><TD>%s:%u</TD><TD>%s:%u</TD><TD>", ts, 
                        inet_ntoa(ts->so_local_addr), ntohs(ts->so_local_port),
                        inet_ntoa(ts->so_remote_addr), ntohs(ts->so_remote_port));
        switch(ts->so_state) {
        case TCPS_LISTEN:
            NutPrintString(sostream, "LISTEN");
            break;
        case TCPS_SYN_SENT:
            NutPrintString(sostream, "SYNSENT");
            break;
        case TCPS_SYN_RECEIVED:
            NutPrintString(sostream, "SYNRCVD");
            break;
        case TCPS_ESTABLISHED:
            NutPrintString(sostream, "<FONT COLOR=#CC0000>ESTABL</FONT>");
            break;
        case TCPS_FIN_WAIT_1:
            NutPrintString(sostream, "FINWAIT1");
            break;
        case TCPS_FIN_WAIT_2:
            NutPrintString(sostream, "FINWAIT2");
            break;
        case TCPS_CLOSE_WAIT:
            NutPrintString(sostream, "CLOSEWAIT");
            break;
        case TCPS_CLOSING:
            NutPrintString(sostream, "CLOSING");
            break;
        case TCPS_LAST_ACK:
            NutPrintString(sostream, "LASTACK");
            break;
        case TCPS_TIME_WAIT:
            NutPrintString(sostream, "TIMEWAIT");
            break;
        case TCPS_CLOSED:
            NutPrintString(sostream, "CLOSED");
            break;
        default:
            NutPrintString(sostream, "?UNK?");
            break;
        }
        NutPrintString(sostream, "</TD></TR>\r\n");
        NutPrintFlush(sostream);
    }

    NutPrintString(sostream, "</TABLE></BODY></HTML>");
    NutPrintFlush(sostream);

    return 0;
}

THREAD(Service, arg)
{
    TCPSOCKET *sock;
    NUTDEVICE *sostream;
    u_char id = (u_char)((u_short)arg);

    /*
     * Now loop endless for connections.
     */
    for(;;) {

        /*
         * Create a socket.
         */
        if((sock = NutTcpCreateSocket()) == 0) {
            NutPrintFormat(0, "[%u] Creating socket failed\r\n", id);
            NutSleep(5000);
            continue;
        }

        /*
         * Listen on port 80. If we return,
         * we got a client.
         */
        NutTcpAccept(sock, 80);
        NutPrintFormat(0, "[%u] Connected, %u bytes free\r\n", id, NutHeapAvailable());

        /*
         * Check for enough memory. If we are below
         * 8 kByte, take a nap and try again.
         */
        for(;;) {
            if(NutHeapAvailable() > 8192)
                break;
            NutPrintFormat(0, "[%u] Low mem\r\n", id);
            NutSleep(1000);
        }

        /*
         * Create a virtual stream device from the socket,
         * so we can use NutPrintFormat() and other stream
         * device goodies.
         */
        if((sostream = NutSoStreamCreate(sock)) == 0) {
            NutPrintFormat(0, "[%u] Creating stream device failed\r\n", id);
        }
        else {
            /*
             * Process http request.
             */
            NutHttpProcessRequest(sostream);

            /*
             * Destroy the virtual stream device.
             */
            NutSoStreamDestroy(sostream);
        }

        /*
         * Close our socket.
         */
        NutTcpCloseSocket(sock);
        NutPrintFormat(0, "[%u] Disconnected\r\n", id);
    }
}

/*!
 * \brief Main application routine. 
 *
 * Nut/OS automatically calls this entry after initialization.
 */
THREAD(NutMain, arg)
{
    u_long baud = 115200;
    u_char i;
	
	/*
	 * Init IDE
	 */
	IDEInit(aSectorBuffer);
	
	/*
	 * Init FAT filesystem
	 */
	FAT32Init();
	 
    /*
     * Initialize the uart device.
     */
    NutDeviceIOCtl(0, UART_SETSPEED, &baud);
    outp(BV(RXEN) | BV(TXEN), UCR);

    NutPrintString(0, "\r\nHTTP Daemon...");
#ifdef NUTDEBUG
    NutKTcpTrace(0);
    NutKOsTrace(1);
    NutKHeapTrace(0);
#endif

    /*
     * Register Realtek controller at address 8300 hex
     * and interrupt 5.
     */
    if(NutRegisterDevice(&devEth0, 0x8300, 5)) 
        NutPrintString(0, "Register failed\r\n");

    /*
     * Configure lan interface. Note that we pass
     * IP-address and IP-mask as zero, which will
     * enable DHCP. If you don't got DHCP, call
     * something like
     */  
    {
        char mac[] = { 0x00, 0x06, 0x98, 0x01, 0x01, 0x09 };
        /*
         * Due to an ICCAVR bug, we need to define extra
         * local variables. Thanks to JianJian, who detected
         * this.
         */
        u_long ip = inet_addr("192.168.1.71");
        u_long msk = inet_addr("255.255.255.0");
        
        NutNetIfConfig("eth0", mac, ip, msk);
    }
    
    /*
    if(NutNetAutoConfig("eth0"))
        NutPrintString(0, "Config failed\r\n");
    */
    NutPrintString(0, "ready\r\n");

    /*
     * Register our CGI sample. This will be called
     * by http://host/cgi-bin/test.cgi?anyparams
     */
    NutRegisterCgi("test.cgi", ShowQuery);

    /*
     * Register more CGI samples.
     */
    NutRegisterCgi("threads.cgi", ShowThreads);
    NutRegisterCgi("timer.cgi", ShowTimer);
    NutRegisterCgi("sockets.cgi", ShowSockets);

    /*
     * Protect the cgi-bin directory with
     * user and password.
     */
    NutRegisterAuth("cgi-bin", "root:root");

    for(i = 1; i <= 4; i++) {
        char *thname = "httpd0";

        thname[5] = '0' + i;
        NutThreadCreate(thname, Service, (void *)(u_short)i, 640);
    }

    /*
     * We could do something useful here, like serving a watchdog.
     */
    NutThreadSetPriority(254);
    for(;;) 
        NutSleep(60000);
}
